% this script implements the SRC method, when there are missing training
% entries
% parameters:
%   dataset:        dataset you want to test, 'AR' or 'YaleB';
%   path:           the path of the dataset;
%   n_People:       number of the subjects in the dataset;
%   n_Condition:    how many images per subjects;
%   height:         the height of image (for resize);
%   width:          the width of image;
%   stage:          the procedure that you want to process, [abc]_2, where
%       a means the training stage, b testing for testing sets and c for
%       testing of testing set;
%   debug:          the control of debug output. when set, information and
%       variable will be saved;
%   limem:          if set, the function works in memory limited mode;
%   mode:           controls the normalization method, default 2:
%       0   :   do nothing;
%       1   :   unit l2 norm for each image;
%       2   :   unit l2 norm and zero mean for each image;
%       3   :   unit std and zero mean for each image;
%   lambda:         for random stream, if specified;
%   tol:            the error tolerance for GPSR;
%   maxIter:        the maximal number of iterations for GPSR;
%   tau:            the weight for sparsity of GPSR;
%   rho:            not used;
%   ratio:          controls the ratio of missing entries or the sampling 
%       matrix;
% =========================================================================
function DemoRandomSRC(dataset, path, n_People, n_Condition, ...
    height, width, stage, debug, limem, mode, ...
    lambda, tol, maxIter, tau, rho, ratio)
%% ========================================================================
% preprocessing stage
% the name of the dataset
% the path of the dataset
% number of people in the database
if nargin<4
    disp('usage: runJointMC.exe dataset path n_People n_Condition');
    exit;
end
option.dataset=dataset;
option.path=path;
if isnumeric(n_People)
    option.n_People=n_People;
else
     option.n_People=str2double(n_People);
end
if isnumeric(n_Condition)
    option.n_Condition=n_Condition;
else
     option.n_Condition=str2double(n_Condition);
end
% optional parameters
if nargin<5 || isempty(height) || isempty(width)
    option.height=0;
    option.width=0;
else
    if isnumeric(height)
        option.height=height;
    else
         option.height=str2double(height);
    end
    if isnumeric(width)
        option.width=width;
    else
         option.width=str2double(width); 
    end
end
% control the process: stage=[abc]2;
% a=1:  run training process
% b=1:  evaluate training data
% c=1:  evaluate testing data
if nargin<7 || isempty(stage)
    option.stage=7;
else
    if isnumeric(stage)
        option.stage=stage;
    else
         option.stage=str2double(stage); 
    end
end
% debug=1: debug mode, all information will be displayed and saved
if nargin<8 || isempty(debug)
    option.debug=0;
else
    if isnumeric(debug)
        option.debug=debug;
    else
         option.debug=str2double(debug); 
    end
end
% limem=1: memory limited mode, temporary variable will be automatically
% cleared. Speed maybe delayed
if nargin<9 || isempty(limem)
    option.limem=1;
else
    if isnumeric(limem)
        option.limem=limem;
    else
         option.limem=str2double(limem); 
    end
end
% normalization method, default 2:
% 0 : do nothing;
% 1 : unit l2 norm for each image;
% 2 : unit l2 norm and zero mean for each image;
% 3 : unit std and zero mean for each image;
if nargin<10 || isempty(mode)
    option.mode=2;
else
    if isnumeric(mode)
        option.mode=mode;
    else
         option.mode=str2double(mode); 
    end
end
% lambda (default 1), controls the weight between low rank and sparse
% regularizer, where large lambad means more sparse result. when Lambda is
% larger than 2.5, there is no difference in the final result
if nargin<11 || isempty(lambda)
    option.lambda=RandStream('mrg32k3a');
else
    option.lambda=lambda;
end
% error tolerance (default 1e-6)
if nargin<12 || isempty(tol)
    option.tol=0.05;
else
    if isnumeric(tol)
        option.tol=tol;
    else
         option.tol=str2double(tol); 
    end
end
% max number of iterations (default 30)
if nargin<13 || isempty(maxIter)
    option.maxIter=1000;
else
    if isnumeric(maxIter)
        option.maxIter=maxIter;
    else
         option.maxIter=str2double(maxIter); 
    end
end
% tau (default 1000 for normalization)
if nargin<14 || isempty(tau)
    option.tau=1000;
else
    if isnumeric(tau)
        option.tau=tau;
    else
         option.tau=str2double(tau); 
    end
end
% rho (not used)
if nargin<15 || isempty(rho)
    option.rho=0;
else
    if isnumeric(rho)
        option.rho=rho;
    else
         option.rho=str2double(rho); 
    end
end
% gross to control whether pack the E_hat into G_hat
if nargin<16 || isempty(ratio)
    option.ratio=0;
else
    if ~isscalar(ratio)
        % load in omega
        omega=ratio;
        option.ratio=sum(ratio(:))/numel(ratio);
    elseif isnumeric(ratio)
        option.ratio=ratio;
    else
        option.ratio=str2double(ratio); 
    end
end
% show the parameter
disp(option);
%% ========================================================================
% training stage
% the random matrix
cd(path);
try
    disp('try to load the random projection');
    load('SRC_sample.mat');
catch
    disp('fail to load the random projection');
    sample=CreateSample(option.width*option.height, option.rho);
    save('SRC_sample.mat', 'sample');
end
if bitand(option.stage, 1)
    % some variable
    disp('training dictionary');
    if ~exist('omega') %#ok<EXIST>
        omega=GenerateSample(option.n_People, option.n_Condition, option.ratio);
    end
    % calculate the condition number, i.e. the ratio of missing entries
    fprintf(1, 'total:\t%.2f\nsubject:\t%.2f\ncondition:\t%.2f', ...
        sum(omega(:))/(option.n_People*option.n_Condition), ...
        min(sum(omega, 2))/option.n_Condition, ...
        min(sum(omega, 1))/option.n_People);
    % calculating the boundary
    bound=zeros(1, option.n_People+1);
    for i=1: option.n_People
        bound(i+1)=bound(i)+sum(omega(i, :));
    end
    % build the dictionary
    dict=TrainSRC(sample, omega, option);
    dict=dict./repmat(sqrt(sum(dict.*dict)), size(dict, 1), 1);
    % to be robust
    save([dataset '-SRC-train.mat'], 'dict', 'omega', 'bound', 'option');
else
    disp('load dictionary');
    load([dataset '-SRC-train.mat'], 'dict', 'bound', 'omega');
end
%% ========================================================================
% start the testing
% for each test image, which should be stored under path\test\
if bitand(option.stage, 2)
    f1=fopen([dataset '-test.txt'], 'w');
    cur_dir=pwd();
    cd([path 'test/']);
    dirlist=dir('*.bmp');
    idx=zeros(1, length(dirlist));
    clear output;
    try
        for i=1:length(dirlist)
            [idx(i), output(i)]=TestSRC(filename, dict, sample, bound, option);
            fprintf(f1, '%s result is %03d\n', dirlist(i).name, idx(i));
        end
        cd(cur_dir);
        save([dataset datestr(now, '-yy-mm-dd-HH-MM-SS-') 'SRC-result.mat'],...
            'idx', 'output', 'option', 'bound', 'omega');
    catch ME
        getReport(ME)
        cd(cur_dir);
        save([dataset '-SRC-exception.mat']);
    end
    fclose(f1);
end
end

%% the training function
function dict=TrainSRC(sample, omega, option)
    dirlist=dir('*.bmp');
    dict=zeros(option.rho, sum(omega(:)));
    k=1;
    for i=1: option.n_People
        for j=1: option.n_Condition
            if omega(i, j)
                img=imread(dirlist((i-1)*option.n_Condition+j).name);
                if size(img, 3)~=1
                    img=rgb2gray(img);
                end
                img=double(img);
                dict(:, k)=sample*img(:);
                k=k+1;
            end
        end
    end
end

%% for testing function
function [idx, output]=TestSRC(filename, dict, sample, bound, option)
    % read in the image
    img=imread(filename);
    if size(img, 3)~=1
        img=rgb2gray(img);
    end
    img=double(img);
    img=sample*img(:);
    % testing
%     output.coef=OMP(dict, img, 2*option.n_Condition);
    x0=dict'*img;
    first_tau_factor = 0.8*(max (max((abs(x0)./option.tau))));
    output.coef=GPSR_BB(img, dict, option.tau,...
         'Debias',0,...
         'Continuation',1,...
         'ContinuationSteps',5,...
         'FirstTauFactor',first_tau_factor,...
         'Monotone',1,...
         'Initialization',x0,...
         'StopCriterion',3,...
         'ToleranceA',option.tol,...
         'Verbose',0);
    output.err=zeros(1, option.n_People);
    for j=1: option.n_People
        output.err(j)=norm(img-dict(:, bound(j)+1: bound(j+1))*...
            output.coef(bound(j)+1: bound(j+1)), 2);
    end
    [val, idx]=min(output.err); %#ok<ASGLU>
end

%% for partion based