% run_JointMC
% this script try to use JSM-MC to do face image classification
% procedure:
% parameters:
%   dataset:        dataset you want to test, 'AR' or 'YaleB';
%   path:           the path of the dataset;
%   n_People:       number of the subjects in the dataset;
%   n_Condition:    how many images per subjects;
%   height:         the height of image (for resize);
%   width:          the width of image;
%   stage:          the procedure that you want to process, [abc]_2, where
%       a means the training stage, b testing for testing sets and c for
%       testing of testing set;
%   debug:          the control of debug output. when set, information and
%       variable will be saved;
%   limem:          if set, the function works in memory limited mode;
%   mode:           controls the normalization method, default 2:
%       0   :   do nothing;
%       1   :   unit l2 norm for each image;
%       2   :   unit l2 norm and zero mean for each image;
%       3   :   unit std and zero mean for each image;
%   lamnda:         the parameters controls the weight between low rank and
%       sparse in the training stage;
%   tol:            the error tolerance during training stage;
%   maxIter:        the maximal number of iterations for training stage;
%   tau:            the step for training stage;
%   rho:            also the step;
%   ratio:          controls the ratio of missing entries;
% =========================================================================
% Example
% run_JointMC('YaleB', '', 38, 16, 48, 42, 3, 0, 1, 2, 2.5)
% run_JointMC('AR', '', 100, 26, 55, 40, 3, 0, 1, 2, 10)
% =========================================================================
% History
% V15:  we use nearest subspace measurement here;
% V12:  we deal with missing entry problem in this version, i.e. the
%   tranining part is based on V5 of TrainJointMC;
% V11:  we use the subspace method, but the subspace is spanned as {X, {A_j}}
%   and {C_i, {A_j}};
% V10:  we use the subspace method to do classification
% V9:   we use TestJointMC method, where the testing image is sparse coded
%   on both A_hat and C_hat. The parameter for GPSR is varied according to
%   the l2 normalization result;
% V8:   we use TestRobustPCA V5 method, where we try to reconstruct the
%   test image by 3 parts. The gross innovation is sum of E_hat
% V7:   deleted;
% V6:   we use randomface with V5;
% V5:   the dictionary is only go through l2 norm normalized without zero
%   mean. Besides, we add a constant atom to the dictionary.
% V4:   we introduce the condition part. The classification is based on SRC
%   with A_hat and C_hat as a dictionary;
% V3:   It modifies the classification part. We add a sparse error term.
%   Then we train a SVM based on the sparse coding result and label of
%   training sets.
% V2:   It uses the gross innovation and C_hat, which is normalized;
% V1:   We use TestJointMC V1, only C_hat and E_hat is used;
%% ========================================================================
function run_JointMC(dataset, path, n_People, n_Condition, ...
    height, width, stage, debug, limem, mode, ...
    lambda, tol, maxIter, tau, rho, ratio)
%% ========================================================================
% preprocessing stage
% the name of the dataset
% the path of the dataset
% number of people in the database
if nargin<4
    disp('usage: runJointMC.exe dataset path n_People n_Condition');
    exit;
end
option.dataset=dataset;
option.path=path;
if isnumeric(n_People)
    option.n_People=n_People;
else
     option.n_People=str2double(n_People);
end
if isnumeric(n_Condition)
    option.n_Condition=n_Condition;
else
     option.n_Condition=str2double(n_Condition);
end
% optional parameters
if nargin<5 || isempty(height) || isempty(width)
    option.height=0;
    option.width=0;
else
    if isnumeric(height)
        option.height=height;
    else
         option.height=str2double(height);
    end
    if isnumeric(width)
        option.width=width;
    else
         option.width=str2double(width); 
    end
end
% control the process: stage=[abc]2;
% a=1:  run training process
% b=1:  evaluate training data
% c=1:  evaluate testing data
if nargin<7 || isempty(stage)
    option.stage=7;
else
    if isnumeric(stage)
        option.stage=stage;
    else
         option.stage=str2double(stage); 
    end
end
% debug=1: debug mode, all information will be displayed and saved
if nargin<8 || isempty(debug)
    option.debug=0;
else
    if isnumeric(debug)
        option.debug=debug;
    else
         option.debug=str2double(debug); 
    end
end
% limem=1: memory limited mode, temporary variable will be automatically
% cleared. Speed maybe delayed
if nargin<9 || isempty(limem)
    option.limem=1;
else
    if isnumeric(limem)
        option.limem=limem;
    else
         option.limem=str2double(limem); 
    end
end
% normalization method, default 2:
% 0 : do nothing;
% 1 : unit l2 norm for each image;
% 2 : unit l2 norm and zero mean for each image;
% 3 : unit std and zero mean for each image;
if nargin<10 || isempty(mode)
    option.mode=2;
else
    if isnumeric(mode)
        option.mode=mode;
    else
         option.mode=str2double(mode); 
    end
end
% lambda (default 1), controls the weight between low rank and sparse
% regularizer, where large lambad means more sparse result. when Lambda is
% larger than 2.5, there is no difference in the final result
if nargin<11 || isempty(lambda)
    option.lambda=2.5;
else
    if isnumeric(lambda)
        option.lambda=lambda;
    else
         option.lambda=str2double(lambda); 
    end
end
% error tolerance (default 1e-6)
if nargin<12 || isempty(tol)
    option.tol=1e-6;
else
    if isnumeric(tol)
        option.tol=tol;
    else
         option.tol=str2double(tol); 
    end
end
% max number of iterations (default 30)
if nargin<13 || isempty(maxIter)
    option.maxIter=100;
else
    if isnumeric(maxIter)
        option.maxIter=maxIter;
    else
         option.maxIter=str2double(maxIter); 
    end
end
% tau (default 1.25)
if nargin<14 || isempty(tau)
    option.tau=1.25;
else
    if isnumeric(tau)
        option.tau=tau;
    else
         option.tau=str2double(tau); 
    end
end
% rho (default 1.5)
if nargin<15 || isempty(rho)
    option.rho=1.5;
else
    if isnumeric(rho)
        option.rho=rho;
    else
         option.rho=str2double(rho); 
    end
end
% gross to control whether pack the E_hat into G_hat
if nargin<16 || isempty(ratio)
    option.ratio=0;
else
    if isnumeric(ratio)
        option.ratio=ratio;
    else
         option.ratio=str2double(ratio); 
    end
end
% show the parameter
disp(option);
%% ========================================================================
% training stage
if bitand(option.stage, 1)
    % training dictionary
    disp('training dictionary');
    % for mdeol 1, we use TrainJoinrMC5
    [C_hat, A_hat, E_hat, output]=TrainJointMC5(dataset, path, option);
    % is a better idea to save the sampling matrix
%     if strcmpi(dataset, 'AR')
%         option.omega=[output(1).omega; output(2).omega];
%     elseif strcmpi(dataset, 'YaleB')
%         option.omega=output.omega;
%     elseif strcmpi(dataset, 'PIE')
%         option.omega=output.omega;
%     end
    option.omega=output(1).omega;
    % calculate the condition number, i.e. the ratio of missing entries
    fprintf(1, 'total:\t%.2f\nsubject:\t%.2f\ncondition:\t%.2f', ...
        2*sum(option.omega(:))/(option.n_People*option.n_Condition), ...
        2*min(sum(option.omega, 2))/option.n_Condition, ...
        min(sum(option.omega, 1))/option.n_People);
    save([dataset '-train.mat'], 'C_hat', 'A_hat', 'E_hat', 'option', 'output');
else
    % we can load in the training data
    disp('load in the training result');
    load([dataset '-train.mat'], 'C_hat', 'A_hat', 'E_hat');
end
disp('No normalization is needed!');
% but we add a zero term to A_hat
A_hat=reshape(A_hat, option.height*option.width, []);
A_hat=[A_hat, zeros(size(A_hat, 1), 1)];
C_hat=reshape(C_hat, option.height*option.width, []);
%% ========================================================================
% start the testing
% for each test image, which should be stored under path\test\
if bitand(option.stage, 2)
    clear output;
    f1=fopen([dataset '-test.txt'], 'w');
    cur_dir=pwd();
    cd([path 'test/']);
    if strcmpi(dataset, 'PIE') || strcmpi(dataset, 'Multi_PIE')
        dirlist=dir('*.jpg');
    else
        dirlist=dir('*.bmp');
    end
    idx=zeros(1, length(dirlist));
    % the end
    try
        tstart=tic;
        for i=1:length(dirlist)
            % so we try E_hat
            % [idx(i), output(i)]=TestSubspace7(dirlist(i).name, ...
            [idx(i), output(i)]=TestSubspace1(dirlist(i).name, ...
                C_hat, A_hat, [], option);
			fprintf(1, '%s result is %03d\n', dirlist(i).name, idx(i));
            fprintf(f1, '%s result is %03d\n', dirlist(i).name, idx(i));
        end
        tstop=toc(tstart);
        disp(['Time for JointMC7 is ', num2str(tstop)]);
        cd(cur_dir);
        save([dataset datestr(now, '-yy-mm-dd-HH-MM-SS-') 'result.mat'],...
            'idx', 'output', 'option');
    catch ME
        getReport(ME)
        cd(cur_dir);
        save([dataset '-exception.mat']);
    end
    fclose(f1);
end
%% ========================================================================
% getting the label of training sets
if bitand(option.stage, 4)
    f1=fopen([dataset '-train.txt'], 'w');
    cur_dir=pwd();
    cd(path);
    if strcmpi(dataset, 'PIE') || strcmpi(dataset, 'Multi_PIE')
        dirlist=dir('*.jpg');
    else
        dirlist=dir('*.bmp');
    end
    idx=zeros(1, length(dirlist));
    try
        % for debug only
        tstart=tic;
        for i=1:length(dirlist)
            % so we try E_hat
            % [idx(i), output(i)]=TestSubspace7(dirlist(i).name, ...
            [idx(i), output(i)]=TestSubspace1(dirlist(i).name, ...
                C_hat, A_hat, [], option);
			fprintf(1, '%s result is %03d\n', dirlist(i).name, idx(i));
            fprintf(f1, '%s result is %03d\n', dirlist(i).name, idx(i));
        end
        tstop=toc(tstart);
        disp(['Time for JointMC7 is ', num2str(tstop)]);
        cd(cur_dir);
        save([dataset '-result.mat'], 'Dict',...
            'idx', 'output', 'option');
    catch ME
        getReport(ME)
        cd(cur_dir);
        save([dataset '-exception.mat']);
    end
    fclose(f1);
end
%% ========================================================================
% learning a MC-SVM on the coefficient of training samples.
% we use the lib-svm here
% not implemented yet.
end

function Z=GenerateGross(X, Y)
    idx=(X.*Y>0);
    Z=zeros(size(X));
    Z(idx)=sign(X(idx)).*max(abs(X(idx)), abs(Y(idx)));
end
